#!/usr/bin/env python3
#
# Created in utilities/file_location_listing#2
#
#
import hashlib
import os
import re
import requests
import sys
import time

from bs4 import BeautifulSoup
from pathlib import Path
from urllib.parse import urlparse
from urllib.parse import urljoin


# Pull in our modules
sys.path.append(os.path.join(os.path.dirname(__file__), '../../lib'))

import storage
import config


# Disable IPV6
requests.packages.urllib3.util.connection.HAS_IPV6 = False

# Suppress cert warnings
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


# Number of pages visited
VISITED_PAGES = 0
MAX_CRAWL_PAGES = 120000
MAX_VISITED_LENGTH = 20000

robots = {}
visitedlinks = {}
linkstoindex = {}
DB_PATH=config.getDBPath()
SITE_LIST = config.getSites()
PATH_HINTS = config.getPathHints()
MY_USER_AGENT = config.getUA()
REINDEX = config.getForceReindex()

# Pass the DB Path into the storage engine
storage.DB_PATH = DB_PATH

# Initiate a session
reqs = requests.session()


def fetchHead(page):
    ''' Do a HEAD so we can check ETag/Last-Mod etc

        Return status code, headers and the final URL if we followed a redirect

    '''
    global reqs
    reqheaders = {"User-agent":MY_USER_AGENT}

    headers = False
    finalurl = page

    try:
        resp = reqs.head(page,headers=reqheaders,allow_redirects=True,verify=False)        
        finalurl = resp.url
        headers = resp.headers
        status = resp.status_code
    except:
        status = 000
        headers = False
        finalurl = page
    
    return [status,headers,finalurl]


def fetchPage(page):
    ''' Simple HTTP fetch of a page
    
    Returns the full response object 
    '''
    
    global reqs
    
    try:
        headers = {"User-agent": MY_USER_AGENT}
        r = reqs.get(page,headers=headers,verify=False)
        return r
    except:
        return False
    

def checkAgainstRobots(url):
    ''' Check a URL against the robots.txt for that domain
    '''
    
    # Extract domain and scheme
    linkparsed = urlparse(url)
    
    # See if we've cached robots.txt for this domain
    global robots
    if linkparsed.netloc not in robots:
        # Nope, so grab it
        robots[linkparsed.netloc] = parseRobotsTxt(linkparsed.scheme,linkparsed.netloc,MY_USER_AGENT)
        
    
    # Now we need to see if the requested path begins with anything in a disallow
    for ban in robots[linkparsed.netloc]['Disallowed']:
        if linkparsed.path.startswith(ban):
            return False
        
    
    # If we got all the way through that without issue, the URL is allowed.
    return True
    

def parseRobotsTxt(scheme,domain,myua=False):
    '''
        Filched and amended from https://stackoverflow.com/a/43086135
    '''    
    print(f"Fetching robots for {scheme}://{domain}/robots.txt")
    result = fetchPage(f"{scheme}://{domain}/robots.txt")
    result_data_set = {"Disallowed":[], "Allowed":[]}
    if not result:
        return result_data_set

    seen_wildcard = False
    
    if check_if_utf8(result.content):
        text = result.content.decode()
    else:
        text = result.content

    for line in text.split("\n"):
        
        # Only pick up on rules for *
        if line.startswith('User-agent'):
            if "*" in line:
                seen_wildcard = True
            elif myua and myua in line:
                seen_wildcard = True
            else:
                seen_wildcard = False
        
        if line.startswith('Allow') and seen_wildcard:    # this is for allowed url
            lp = line.split(': ')
            if len(lp) == 1 :
                # It's Disallow: on it's own on a line
                continue            
            result_data_set["Allowed"].append(lp[1].split(' ')[0])    # to neglect the comments or other junk info
        elif line.startswith('Disallow') and seen_wildcard:    # this is for disallowed url
            lp = line.split(': ')
            if len(lp) == 1 :
                # It's Disallow: on it's own on a line
                continue
            result_data_set["Disallowed"].append(lp[1].split(' ')[0])    # to neglect the comments or other junk info

    return result_data_set

def check_if_utf8(s):
    ''' Some responses will already be UTF-8
    Try and decode them to identify whether this is the case
    '''
    try:
       s.decode('utf-8')
       return True
    except (UnicodeError, AttributeError):
       is_utf8 = False
 
    try:
       s.decode('latin1')
       return True
    except (UnicodeError, AttributeError):
       is_utf8 = False
 
    return is_utf8

def createLinkObject():
    ''' Return a dict in the schema used by the crawler when queuing links to index
    '''
    linkinfo = {
            "href": "",
            "text": [],
            "linktype": "external",
            "visited": False,
            "domain": "",
            "is_embed": False
    }
    
    return linkinfo

def tidyVisited():
    '''
        Prevent the global visitedlinks from getting too large
        I left one property indexing overnight, and found the crawler using 57% of available RAM because of the number of URLs it had crawled
    '''
    global visitedlinks
    
    # Get the size of the dict
    dictlen = len(visitedlinks)
    
    # Is it greater than the configured limit?
    if dictlen > MAX_VISITED_LENGTH:

        # Iterate over the dict, deleting keys until we're under the limit, allowing a 25% margin
        # otherwise we'd be running this every cycle
        x = dictlen
        d = []
        for key in visitedlinks:
            d.append(key)
            x=x-1
            if x < (MAX_VISITED_LENGTH * 0.75):
                break
            
        for key in d:
            del(visitedlinks[key])


def extractUrlsFromHTML(pagecontent, url):
    ''' Extract title and outgoing links
    '''
    
    title = False
    outgoinglinks = {}
    
    if pagecontent.title:
        title = pagecontent.title.string    
        
    links = pagecontent.findAll('a')
    for link in links:
        linkinfo = createLinkObject()
        
        if link.has_attr('rel') and link['rel'] == "nofollow":
                continue

        # Get the link target
        if not link.has_attr('href'):
                continue

        try:
            if link['href'] == "../":
                    continue
        except:
            # It's a named anchor
            continue
        
        if "javascript:" in link['href']:
                continue
            
        if "_javascript_:" in link['href']:
                continue
                        
        if "mailto:" in link['href']:
                continue

        if "abp:subscribe" in link['href']:
                continue

        linkinfo['href'] = link['href']

        # Parse the destination link
        linkparsed = urlparse(link['href'])

        if link.has_attr('rel'):
                linkinfo['rel'] = link['rel']

        if not linkparsed.path and linkparsed.fragment:
                # Link is to a fragment within the page
                continue

        if not linkparsed.netloc:
                # Link is within the same domain
                linkinfo['href'] = urljoin(url,link['href'])
        else:
                linkinfo['domain'] = linkparsed.netloc

        # If it has a title, make a note
        if link.has_attr('title') and len(link['title']) > 0:
                linkinfo['title'] = link['title']

        if link.text:
                linkinfo['text'] = [link.text]

        # Strip any fragment
        p = linkinfo['href'].split("#")
        linkinfo['href'] = p[0]
        
        # Don't queue it if we wouldn't be allowed to crawl it
        if not shouldCrawlURL(linkinfo['href'], quiet=True):
            continue

        if linkinfo['href'] not in outgoinglinks:
                outgoinglinks[linkinfo['href']] = linkinfo
        elif link.text:
                outgoinglinks[linkinfo['href']]['text'].append(link.text)


    # Check for any image embeds
    images = pagecontent.findAll('img')
    for image in images:
        linkinfo = createLinkObject()
        if not image.has_attr('src'):
                continue
    
        linkinfo['href'] = image['src']
        linkinfo['linked_from'] = url
        linkparsed = urlparse(linkinfo['href'])
        
        if not linkparsed.netloc:
                # Link is within the same domain
                linkinfo['href'] = urljoin(url,image['src'])
                linkinfo['linktype'] = 'internal'
        else:
                linkinfo['domain'] = linkparsed.netloc


        if image.has_attr('alt') and len(image['alt']) > 0:
                linkinfo['title'] = image['alt']
        else:
                linkinfo['title'] = title
                
        # Don't queue it if we wouldn't be allowed to crawl it
        if not shouldCrawlURL(linkinfo['href'], quiet=True):
            continue
        
        linkinfo['is_embed'] = True
        if linkinfo['href'] not in outgoinglinks:
                outgoinglinks[linkinfo['href']] = linkinfo        

    return outgoinglinks, title
    
def extractUrlsFromText(plaintext):
    ''' Regex out any URLs in a block of text
    '''
    try:
        urls = re.findall("(https?://[^\s^\]^\)]+)", plaintext.decode())
        outgoinglinks = {}
        for link in urls:
            
            if not shouldCrawlURL(url, quiet=True):
                continue
            
            linkinfo = createLinkObject()
            linkinfo['href'] = link
            linkparsed = urlparse(linkinfo['href'])
            linkinfo['domain'] = linkparsed.netloc

            if linkinfo['href'] not in outgoinglinks:
                    outgoinglinks[linkinfo['href']] = linkinfo

        return outgoinglinks    
    except:
        return {}
    
    
def createVisitedHash(url):
    ''' Length of an average mailarchive URL is 80 bytes. 
        Using a MD5 reduces it to 32, which means we need less memory to store 
        the list of what we've visited, at the cost of slightly more CPU usage
    '''
    
    if check_if_utf8(url):
        return hashlib.md5(url).hexdigest()
    else:
        return hashlib.md5(url.encode('utf-8',errors="replace")).hexdigest()    


def shouldCrawlURL(url, quiet=False):
    ''' Check whether we're supposed to crawl the supplied URL
    '''
    
    # Is the link to a domain we're allowed to crawl?
    #
    # Should probably also look at checking this when
    # queueing
    parsed = urlparse(url)
    parsed_host = f"{parsed.scheme}://{parsed.netloc}"

    if parsed_host.lower() not in SITE_LIST:
        if not quiet:
            print(f"Not on an authorised domain {url}")
        storage.deleteStoredItem(url)
        return False    

    if config.shouldSkipURL(url, parsed.netloc):
        visitedlinks[createVisitedHash(url)] = "skip"
        if not quiet:
            print(f"Skip-rule match for {url}")
        storage.deleteStoredItem(url)
        return False
                
    if not checkAgainstRobots(url):
        # URL is prohibited by the domain's robots.txt
        visitedlinks[createVisitedHash(url)] = "d"
        if not quiet:
            print(f"Skipping {url} due to robots.txt")
        storage.deleteStoredItem(url)
        return False
    
    return True
    
    
    

def crawlPage(url, override = False, ele = False):
    ''' Crawl a URL and store the result
    
    
    '''

    # Fetch config
    global visitedlinks
    global MAX_CRAWL_PAGES
    global VISITED_PAGES
    global linkstoindex

    if VISITED_PAGES > MAX_CRAWL_PAGES:
        print("Reached MAX_CRAWL_PAGES. Aborting")
        return True
    
    # Check if we're supposed to crawl
    if not shouldCrawlURL(url):
        return False
    
    print(f"Indexing {url}")
    if not override and not url.endswith("/"):
        last_mod, content_hash, valid_date = storage.getStoredInfo(url)
    else:
        last_mod = 'FFFFFF'
        content_hash = 'FFFFFF'
        valid_date = 'FFFFFF'    

    # Check whether we've already indexed it today
    if valid_date == time.strftime('%Y-%m-%d') and not REINDEX:
        print("\tPage already indexed today. Skipping")
        visitedlinks[createVisitedHash(url)] = "d"
        return True

    # Otherwise fetch and process it
    try:
        page_head = fetchHead(url)
        #print(page_head)
        VISITED_PAGES += 1
        
        # Check whether we were redirected
        if page_head[2] != url:
            visitedlinks[createVisitedHash(url)] = "d"
            storage.deleteStoredItem(url)
            url = page_head[2]
        
        # Check the status code
        if page_head[0] != 200:
            print(f"Non-200 response to HEAD for {url}")
            # TODO - try a GET, some servers don't do HEAD
            storage.deleteStoredItem(url)
            return False
        
        # Check whether last modified is present and matches, if so, we can skip
        if (page_head[1].get("Last-Modified") is None or page_head[1].get("Last-Modified") != last_mod):
            # We have 2 things to do now then
            #
            # Fetch the URL
            # (if it's an appropriate content-type) search it for other URLs
            # Build an index entry for this URL
            pd = storage.buildPageDict(url, page_head[1])
            #print(pd)
            
            # If it's a filetype we're interested in
            # Fetch the page itself
            pg = False
            fetch_types = [
                "text/plain",
                "text/html",
                "image/jpeg",
                "image/png",
                "image/gif"
                ]
            if pd["content-type"] in fetch_types:
                pg = fetchPage(url)
            
            # Initilise some variables
            title = False
            links = False
            
            if pd["bytes"] is False and pg:
                print("\tCalculating page length")
                pd["bytes"] = len(pg.content)
            
            # Check the content-type, should we be checking the response for links to other places?
            if pd["content-type"] == "text/plain":                    
                links = extractUrlsFromText(pg.content)
                if url.endswith(".md") or url.endswith(".md.txt"):
                    # The first line should contain the title
                    if check_if_utf8(pg.content):
                        lines = pg.content.decode().split("\n")
                    else:
                        lines = pg.content.split("\n")
                    if lines[0].startswith('#') or lines[1].startswith("---") or lines[1].startswith("==="):
                        title = lines[0].strip("#").strip()
                
            if pd["content-type"] == "text/html":
                pagecontent = BeautifulSoup(pg.content, features="lxml")
                links, title = extractUrlsFromHTML(pagecontent, url)

            if pd["content-type"] in [ "image/jpeg", "image/png", "image/gif"]:
                pd['image-path'] = storage.storeImage(url, pg, pd["content-type"])

            # If we've been passed parent element info, this is probably
            # an image, so we should take the title from the tag that embedded
            # it
            if ele and ele['is_embed']:
                pd['title'] = ele['title']
                if "linked_from" in ele:
                    pd['embedded-in'] = ele['linked_from']

            if title:
                pd['title'] = title
                
            # Build the index entry
            storage.storeInfo(pd)
            
            # Mark the URL as visited
            visitedlinks[createVisitedHash(url)] = "d"
            
            # if we found outbound links, queue them
            if links:
                linkstoindex.update(links)
      
        else:
            print("\tUnchanged")
        

    except KeyboardInterrupt:
        raise

    except Exception as e:
        print("Indexing page failed")
        print(e)

    return True

def triggerRecurse():
    ''' Recurse through the crawl queue until everything's been visited
    '''
    global visitedlinks
    global linkstoindex
    global VISITED_PAGES
    
    while True:
        if VISITED_PAGES > MAX_CRAWL_PAGES:
            break
        
        if len(linkstoindex) < 1:
            print("No more links")
            break
        
        # Get the key and make a shallow copy
        last_key = list(linkstoindex.keys())[-1]
        ele = dict(linkstoindex[last_key])
        
        # remove it from the queue
        del(linkstoindex[last_key])
        
        # Hash it and check whether we visited when
        # queuing
        urlhash = createVisitedHash(ele['href'])
        if ele['visited']:
            visitedlinks[urlhash] = "done"
            continue
        
        # Check whether we visited already
        if urlhash in visitedlinks:
            continue
        
        crawlPage(ele['href'], False, ele)
        
        tidyVisited()



### Storage


if __name__ == "__main__":        
    
    if len(sys.argv) > 1:
        crawlPage(sys.argv[1])
    else:
        # Crawl the first page of each site
        for site in SITE_LIST:
            crawlPage(site, override = True)

    # Queue up any paths provided via hint
    hinted_paths = {}
    for hint in PATH_HINTS:
        if shouldCrawlURL(hint, quiet=False):
            linkinfo = createLinkObject()
            linkinfo['href'] = hint
            linkparsed = urlparse(linkinfo['href'])
            linkinfo['domain'] = linkparsed.netloc

            if linkinfo['href'] not in hinted_paths:
                    hinted_paths[linkinfo['href']] = linkinfo
    
    linkstoindex.update(hinted_paths)
                
    # We may have queued new links, so trigger recursion
    triggerRecurse()

    # Rebuild the index
    storage.indexStorage()
