#!/usr/bin/env python3
#
# Storage module
#

import hashlib
import gzip
import json
import os
import sys
import tempfile
import time

from pathlib import Path
from subprocess import Popen
from urllib.parse import urlparse

sys.path.append(os.path.join(os.path.dirname(__file__), '../lib'))

import config


INDEXES = {
    "ALL" : [],
    "IMAGE" : [],
    "DOC" : [],
}
DB_PATH = '/opt/search_db'
INDEX_ALL = False
LOADED_AT = False
CONCURRENCY = config.getConcurrency()


def loadIndex():
    ''' Load the index from disk
    
    '''
    
    global LOADED_AT
    global INDEXES
    
    # Do we already have an index loaded in memory?
    if len(INDEXES["IMAGE"]) > 0 or len(INDEXES["ALL"]) > 0 or len(INDEXES["DOC"]) > 0:
        # Blank it
        INDEXES = {
            "ALL" : [],
            "IMAGE" : [],
            "DOC" : [],
        }
        
    tombstones = {}
    if os.path.exists(f"{DB_PATH}/index.tomb"):
        with open(f"{DB_PATH}/index.tomb", 'r') as f:
            for l in f.read().splitlines():
                tombstones[l.strip()] = 1
    
    with open(f"{DB_PATH}/index", 'r') as f:
          # We mark each entry with an index number
          # this allows the search to parallelise by having
          # threads only consider entries with their corresponding
          # index number
          idx_num = 0
          
          # Iterate through the loaded file
          for l in f.read().splitlines():
            index_l = l.split(",!,")
            
            # Has the URL been tombstoned?
            if index_l[0] in tombstones:
                continue
            
            # Should we be populating the "ALL" index
            #
            # This is off by default - generally speaking we
            # tend to want to search by type, so having ALL is
            # just wasting RAM
            if INDEX_ALL:
                INDEXES["ALL"].append({
                    "u" : index_l[0], # url
                    "h" : index_l[1], # index hash
                    "t" : index_l[2], # content-type
                    "n" : index_l[3], # title/name
                    "i" : idx_num, # Index number
                    })
                
            if index_l[2] in ["image/jpeg", "image/png", "image/gif"]:
                INDEXES["IMAGE"].append({
                    "u" : index_l[0], # url
                    "h" : index_l[1], # index hash
                    "t" : index_l[2], # content-type
                    "n" : index_l[3], # title/name
                    "i" : idx_num, # Index number
                    })
            else:
                INDEXES["DOC"].append({
                    "u" : index_l[0], # url
                    "h" : index_l[1], # index hash
                    "t" : index_l[2], # content-type
                    "n" : index_l[3], # title/name 
                    "i" : idx_num, # Index number
                    })
                
            idx_num += 1
            if idx_num >= CONCURRENCY:
                idx_num = 0

    LOADED_AT = time.time()
    INDEXES["last_mod"] = getIndexLastMod()


def deleteStoredItem(url):
    '''
    Return stored item for the url
    '''
    
    # calculate a hash of the URL
    h = hashlib.sha256()
    h.update(url.encode('utf-8'))
    fname = h.hexdigest()
    
    # Calculate dir path
    dir_path = getStoragePath(fname)
    
    file_path = f"{dir_path}/{fname}"
    
    if not os.path.exists(file_path):
        return False
    
    #print(f"Will delete{file_path}")
    os.remove(file_path)
    
    # Check if that's resulted in an empty directory
    if len(os.listdir(dir_path)) == 0:
        try:
            os.rmdir(dir_path)
        except:
            # oh well, no harm done
            pass

    # Also need to check whether there's a thumbnail
    thumbs_path = getThumbsPath(fname)
    path_prefix = f"{thumbs_path}/{fname}"
    for ext in [".gif", ".jpg"]:
        file_path = f"{path_prefix}{ext}"
        if not os.path.exists(file_path):
            continue
        
        # Otherwise, remove it
        os.remove(file_path)
        
    # Tombstone the URL so it'll be ignored in the index
    # at next index reload
    with open(f"{DB_PATH}/index.tomb", 'a+') as f:
        f.write(f"{url}\n")
        
    
def getIndexLastMod():
    ''' Get the mtime of the index
    If there's a tombstone **and** it's more recent, that
    time is returned instead
    '''
    
    tomb = False
    if os.path.exists(f"{DB_PATH}/index.tomb"):
        tomb = os.path.getmtime(f"{DB_PATH}/index.tomb")
        
    idx = os.path.getmtime(f"{DB_PATH}/index")
    
    if tomb and tomb > idx:
        return tomb
    else:
        return idx
    

def getIsIndexFresh():
    ''' Check whether the index has changed since it was loaded
    
    This is achieved by getting mtime and comparing to the mtime
    stored with the index
    '''
    return (getIndexLastMod() == INDEXES['last_mod'])

def getStoredInfo(url):
    # Return stored info for the url
    # TODO
    
    # calculate a hash of the URL
    h = hashlib.sha256()
    h.update(url.encode('utf-8'))
    fname = h.hexdigest()
    
    # Calculate dir path
    dir_path = getStoragePath(fname)
    file_path = f"{dir_path}/{fname}"
    
    j = readStoredFile(file_path)
    if not j:
        return 'FFFFFF', 'FFFFFF', 'FFFFFF'
    
    # return
    # last_mod, content_hash, valid_date
    # Support for content hashing isn't implemented yet
    return j['last-mod'], 'FFFFFFFFF', j['valid-at']


def getStoredItem(url, fname=False):
    '''
    Return stored item for the url
    
    If a fname is provided, that's used instead
    of calculating from the URL
    '''
    
    if not fname:
        # calculate a hash of the URL
        h = hashlib.sha256()
        h.update(url.encode('utf-8'))
        fname = h.hexdigest()
    
    # Calculate dir path
    dir_path = getStoragePath(fname)
    file_path = f"{dir_path}/{fname}"

    # read the file and return the JSON
    return readStoredFile(file_path)


def readStoredFile(file_path):
    ''' Open a storage file for reading and return the json
    
    Attempts to read the file using gzip (the new storage format)
    and falls back to plain (the old) if that fails
    
    The fallback will eventually be removed.
    '''
    
    if not os.path.exists(file_path):
        return False
    
    with gzip.open(file_path) as f:
        for line in f:
            pass
        
        # line is now the last line
        try:
            j = json.loads(line)
        except:
            print("Failed to load file")
            invalidateFile(file_path)
            return False    
        
    return j   


def getStoragePath(fname, full=True):
    ''' Calculate the directory path
    
    '''
    parent_dir1 = fname[0:2]
    parent_dir2 = fname[2:4]
    if full:
        return f"{DB_PATH}/store/{parent_dir1}/{parent_dir2}"
    else:
        return f"{parent_dir1}/{parent_dir2}"


def getThumbsPath(fname, full=True):
    ''' Calculate the directory path for a thumbnail
    
    '''
    parent_dir1 = fname[0:2]
    parent_dir2 = fname[2:4]
    if full:
        return f"{DB_PATH}/thumbs/{parent_dir1}/{parent_dir2}"
    else:
        return f"{parent_dir1}/{parent_dir2}"   
    

def invalidateFile(filepath):
    ''' Mark a file as being invalid
    '''
    print(f"Marking as invalid: {filepath}")
    try:
        os.replace(filepath, f"{filepath}.invalid")    
    except Exception as e:
        print(f"Unable to mark invalid: {e}")
    

def indexStorage():
    ''' After a run, we need to rebuild the index
    
    This is essentially just an ordered list of URLs
    '''
    index_list = []
    for subdir, dirs, files in os.walk(f"{DB_PATH}/store"):
        for file in files:
            # Skip temporary and invalid files
            if file.endswith(".invalid") or file.endswith(".tmp"):
                continue
            
            filepath = os.path.join(subdir, file)
            entry = [False, False, False, False]
            # Open the file and extract the key and hash
            try:
                with gzip.open(filepath, 'r') as f:
                    for line in f:
                        line = line.decode()
                        if line.startswith("KEY"):
                            entry[0] = ":".join(line.split(":")[1:]).strip()
                        if line.startswith("HASH"):
                            entry[1] = ":".join(line.split(":")[1:]).strip()
                        if line.startswith("CONTENTTYPE"):
                            entry[2] = ":".join(line.split(":")[1:]).strip()
                        if line.startswith("TITLE"):
                            entry[3] = ":".join(line.split(":")[1:]).strip()
            except:
                invalidateFile(filepath)
                            
            if entry[2] is False:
                entry[2] = "!unk!"
                
            if entry[3] is False:
                entry[3] = "!unk!"
                
            # We should now have an entry with at least 2 parts. If we don't
            # skip this file
            if len(entry) < 4 or False in entry:
                continue
            
            # The index entry is:
            # url, path
            # Calculate the path
            pth = getStoragePath(entry[1], False)
            index_list.append(f"{entry[0]},!,{entry[1]},!,{entry[2]},!,{entry[3]}")
            
    # Sort the list
    index_list.sort()
    
    # Write to the index
    with open(f"{DB_PATH}/index.tmp", 'w') as f:
        f.write('\n'.join(index_list))
        
    # Put the index into use 
    os.replace(f"{DB_PATH}/index.tmp", f"{DB_PATH}/index")
    
    # Remove the tombstone file, if it exists
    if os.path.exists(f"{DB_PATH}/index.tomb"):
        os.remove(f"{DB_PATH}/index.tomb")
        


def storeInfo(pd):
    # Write out to storage 
    # TODO
    # For now, this just prints
    
    file_str = (
        f"KEY:{pd['key']}\n"
        f"HASH:{pd['hashkey']}\n"
        f"DOMAIN:{pd['domain']}\n"
        f"PATH:{pd['path']}\n"
        f"FILE:{pd['filename']}\n"
        f"CONTENTTYPE:{pd['content-type']}\n"
        f"TITLE:{pd['title']}\n"
        "\n"
    )
    file_str += json.dumps(pd)
    
    #print("===== ENTRY =====")
    #print(file_str)
    #print("===== ENTRY ENDS =====")
    
    # Store to disk
    dir_path = getStoragePath(pd['hashkey'])
    
    # ensure it exists
    path = Path(dir_path)
    path.mkdir(parents=True, exist_ok=True)
    
    # Open the file for writing
    fh = gzip.open(f"{dir_path}/{pd['hashkey']}.tmp", 'wt')
    fh.write(file_str)
    fh.close()
    
    # Move into place
    os.replace(f"{dir_path}/{pd['hashkey']}.tmp", f"{dir_path}/{pd['hashkey']}")
    
    
def storeImage(url, pg, contenttype):
    ''' We've fetched an image
    
    Thumbnail it
    '''
    h = hashlib.sha256()
    h.update(url.encode('utf-8'))
    fname = h.hexdigest()
    
    thumbs_path = getThumbsPath(fname, full=True)
    
    # ensure it exists
    path = Path(thumbs_path)
    path.mkdir(parents=True, exist_ok=True)    
    
    # Save the main image into a temporary file
    tf = tempfile.NamedTemporaryFile(delete=True)
    tf.write(pg.content)
    tf.seek(0)
    
    if contenttype in ["image/gif"]:
        print(f"Thumbnailing gif {url}")
        # Can't thumnail a gif into a jpg, you just get lots of individual frames
        # Thumbnail it
        out, err = Popen(["convert",tf.name,"-loop", "0", "-resize","x450","-quality","75",f"{thumbs_path}/{fname}.tmp.gif"]).communicate()
        tf.close()   
        os.replace(f"{thumbs_path}/{fname}.tmp.gif", f"{thumbs_path}/{fname}.gif")
        
    else:
        # Thumbnail it
        out, err = Popen(["convert",tf.name,"-resize","x450","-quality","75",f"{thumbs_path}/{fname}.tmp.jpg"]).communicate()
        tf.close()   
        os.replace(f"{thumbs_path}/{fname}.tmp.jpg", f"{thumbs_path}/{fname}.jpg")
    
    return getThumbsPath(fname, full=False)
    

def buildPageDict(url, headers):
    ''' Take URL etc and create a dict using a standardised format
    
    This function is the authority on the structure of that format
    '''
    r = {
        "key" : url,
        "domain" : "",
        "path" : "",
        "filename" : "",
        "bytes" : False,
        "last-mod" : "FFFFFF",
        "content-type" : "FFFFFF",
        "title" : "",
        "valid-at" : time.strftime('%Y-%m-%d'),
        "image-path" : False
        }
    
    if headers.get("Last-Modified") is not None:
        r["last-mod"] = headers.get("Last-Modified")
        
    if headers.get("Content-Type") is not None:
        r["content-type"] = headers.get("content-type").lower()
        
    if headers.get("Content-Length") is not None:
        r["bytes"] = headers.get("Content-Length")
    
    
    h = hashlib.sha256()
    h.update(url.encode('utf-8'))
    r["hashkey"] = h.hexdigest()
    
    linkparsed = urlparse(url)
    r["domain"] = linkparsed.netloc
    r["path"] = linkparsed.path
    r["filename"] = r["path"].split("/")[-1]
    
    return r

